from collections import defaultdict

from cemento.draw_io.diagram_ref import DiagramRef
from cemento.draw_io.read_diagram import ReadDiagram


class ReadAreaDiagram(ReadDiagram):

    def __init__(
        self,
        file_path,
        do_check_errors=False,
        diagram_ref=None,
        inverted_rank_arrows=False,
        parse_terms=False,
    ):
        if diagram_ref is None:
            diagram_ref = DiagramRef()

        self.set_file_path(file_path)
        self.set_diagram_ref(diagram_ref)
        self._set_inverted_rank_arrows(inverted_rank_arrows)
        self._set_check_errors(do_check_errors)

        self._elements = self._parse_elements()
        self._edges, self._terms, self._rels = self._read_edges(
            elements=self._get_elements(),
            return_entity_ids=True,
        )
        super().__init__(
            file_path,
            do_check_errors=do_check_errors,
            diagram_ref=diagram_ref,
            inverted_rank_arrows=inverted_rank_arrows,
            parse_terms=parse_terms,
            edges=self._get_edges(),
        )

        self._area_designations = None
        self._area_terms = None

        self._set_area_terms()
        self._set_node_areas()

    def _set_node_areas(self):
        areas = dict()
        terms = set(self._get_terms())
        elements = self._get_elements()

        area_terms = self.get_area_terms().keys()

        for area_term in area_terms:
            x_0, y_0 = int(float(elements[area_term]["x"])), int(
                float(elements[area_term]["y"])
            )
            x_1, y_1 = (
                x_0 + int(float(elements[area_term]["width"])),
                y_0 + int(float(elements[area_term]["height"])),
            )
            areas[area_term] = (x_0, y_0, x_1, y_1)
        # check if node_terms belong to any of the areas
        node_terms = terms - area_terms
        area_designations = defaultdict(set)
        # classify term areas
        for node_term_id in node_terms:
            pos_x, pos_y = (
                int(float(elements[node_term_id].get("x", 0))),
                int(float(elements[node_term_id].get("y", 0))),
            )
            for area_term_id, area in areas.items():
                x_0, y_0, x_1, y_1 = area
                if pos_x > x_0 and pos_x < x_1 and pos_y > y_0 and pos_y < y_1:
                    area_designations[node_term_id].add(area_term_id)
        # classify containing area of areas
        for area_term in area_terms:
            curr_x_0, curr_y_0, curr_x_1, curr_y_1 = areas[area_term]

            for comp_area in area_terms:
                if comp_area == area_term:
                    continue

                comp_x_0, comp_y_0, comp_x_1, comp_y_1 = areas[comp_area]
                if (
                    comp_x_0 < curr_x_0
                    and comp_y_0 < curr_y_0
                    and comp_x_1 > curr_x_1
                    and comp_y_1 > curr_y_1
                ):
                    area_designations[area_term].add(comp_area)

        self._area_designations = area_designations

    def _parse_designations(self, designations):
        elements = self._get_elements()
        parsed_dict = dict()

        for key, values in designations.items():
            new_values = set()
            for value in values:
                new_values.add((value, ReadDiagram.clean_term(elements[value]['value'])))
            new_key = (key, ReadDiagram.clean_term(elements[key]['value']))
            parsed_dict[new_key] = new_values
        return parsed_dict

    def get_area_designations(self, parse_values=False):
        if not parse_values:
            return self._area_designations

        return self._parse_designations(self._area_designations)

    def get_node_designations(self, parse_values=False):
        inverted_dict = defaultdict(set)
        for key, values in self._area_designations.items():
            for value in values:
                inverted_dict[value].add(key)

        if not parse_values:
            return inverted_dict

        return self._parse_designations(inverted_dict)

    def _get_elements(self):
        return self._elements

    def _get_edges(self):
        return self._edges

    def _get_terms(self):
        return self._terms

    def _get_rels(self):
        return self._rels

    def _set_area_terms(self):
        elements = self._get_elements()
        terms = self._get_terms()

        area_terms = {
            term_id: elements[term_id]
            for term_id in terms
            if "~" in elements[term_id]["value"]
        }

        self._area_terms = area_terms

    def get_area_terms(self):
        return self._area_terms



from os import path
import uuid
from datetime import datetime, timezone

from cemento.draw_io.diagram import Diagram
from cemento.draw_io.diagram_ref import DiagramRef

class Shape:

    def __init__(self, id, content, x_pos, y_pos, width, height, ref):
        self._id = id
        self._content = content
        self._x_pos = x_pos
        self._y_pos = y_pos
        self._width = width
        self._height = height
        self._ref = ref
        self._prefix = None
        self._template_key = 'shape'

        self._parse_prefix()

    def _parse_prefix(self):
        try:
            term_sep = self._ref.get_shape_param('term_separator')
            prefix, term_name = self._content.split(term_sep)
            self._prefix = prefix
        except ValueError:
            self._prefix = None
        except AttributeError:
            self._prefix = None

    def _get_fill_color(self):
        try:
            return self._ref.get_term_color(self.get_prefix())
        except KeyError:
            return self._ref.get_term_color('default')

    def get_prefix(self):
        return self._prefix

    def get_id(self):
        return self._id

    def get_content(self):
        return self._content

    def get_x_pos(self):
        return self._x_pos

    def get_y_pos(self):
        return self._y_pos

    def get_width(self):
        return self._width

    def get_height(self):
        return self._height

    def get_ref(self):
        return self._ref

    def set_ref(self, ref):
        self._ref = ref

    def get_key(self):
        return self._template_key

    def get_props(self):
        return {
            'shape_id': self.get_id(),
            'shape_content': self.get_content(),
            'fill_color': self._get_fill_color(),
            'x_pos': self.get_x_pos(),
            'y_pos': self.get_y_pos(),
            'shape_width': self.get_width(),
            'shape_height': self.get_height()
        }

class Circle(Shape):

    def __init__(self, id, content, x_pos, y_pos, radius, ref):
        super().__init__(id, content, x_pos, y_pos, radius, radius, ref)
        self._radius = radius
        self._template_key = 'circle'

class Connector:

    def __init__(self, id, source_id, target_id, connector_label_id, connector_content, is_rank=None, start_x=None, start_y=None, end_x=None, end_y=None, inverted=False):
        self._id = id
        self._source_id = source_id
        self._target_id = target_id
        self._connector_label_id = connector_label_id
        self._connector_content = connector_content
        self._is_rank = is_rank
        self._start_pos_x = start_x
        self._start_pos_y = start_y
        self._end_pos_x = end_x
        self._end_pos_y = end_y
        self._inverted = inverted
        self._template_key = 'connector'

        if is_rank is None:
            self._is_rank = False

        self._is_dashed = 0 if self.get_is_rank() else 1
        self._is_curved = 0 if self.get_is_rank() else 1

        if not(start_x and start_y and end_x and end_y):
            self._update_start_end_pos()

    def _update_start_end_pos(self):
            if self.get_is_rank() and self.get_is_inverted():
                self._start_pos_x = 1
                self._end_pos_x = 0
                self._start_pos_y = 0.5
                self._end_pos_y = 0.5

            if not self.get_is_rank():
                self._start_pos_x = 0.5
                self._end_pos_x = 0.5
                self._start_pos_y = 0.5
                self._end_pos_y = 0.5

            if self.get_is_rank() and not self.get_is_inverted():
                self._start_pos_x = 0.5
                self._end_pos_x = 0.5
                self._start_pos_y = 1
                self._end_pos_y = 0

    def get_id(self):
        return self._id

    def get_source_id(self):
        return self._source_id

    def get_target_id(self):
        return self._target_id

    def get_connector_label_id(self):
        return self._connector_label_id

    def get_connector_content(self):
        return self._connector_content

    def get_is_inverted(self):
        return self._inverted

    def get_is_rank(self):
        if self._is_rank is None:
            raise ValueError("Cannot use rank inference features. Please set is_rank or resort to manual connector positions.")
        return self._is_rank

    def get_start_pos_x(self):
        if self._start_pos_x is None:
            raise ValueError("Cannot access start and end positions. Use the is_rank argument or resort to manual connector positions.")
        return self._start_pos_x

    def get_start_pos_y(self):
        return self._start_pos_y

    def get_end_pos_x(self):
        return self._end_pos_x

    def get_end_pos_y(self):
        return self._end_pos_y

    def get_key(self):
        return self._template_key

    def _get_is_dashed(self):
        return self._is_dashed

    def _get_is_curved(self):
        return self._is_curved

    def get_props(self):
        # if the connection is a rank edge, invert the arrow position according to convention
        source_id = self.get_source_id() if self.get_is_rank() else self.get_target_id()
        target_id = self.get_target_id() if self.get_is_rank() else self.get_source_id()

        return {
            'connector_id': self.get_id(),
            'start_pos_x': self.get_start_pos_x(),
            'start_pos_y': self.get_start_pos_y(),
            'end_pos_x': self.get_end_pos_x(),
            'end_pos_y': self.get_end_pos_y(),
            'source_id': source_id,
            'target_id': target_id,
            'rel_x_pos': 0,
            'rel_y_pos': 0,
            'connector_label_id': self.get_connector_label_id(),
            'connector_val': self.get_connector_content(),
            'is_dashed': self._get_is_dashed(),
            'is_curved': self._get_is_curved(),
        }

class StraightConnector(Connector):

    def __init__(self, id, source_id, target_id, connector_label_id, connector_content, start_x=None, start_y=None, end_x=None, end_y=None):
        super().__init__(id, source_id, target_id, connector_label_id, connector_content, is_rank=False, start_x=start_x, start_y=start_y, end_x=end_x, end_y=end_y, inverted=False)
        self._template_key = "straight_connector"

class WriteDiagram(Diagram):

    def __init__(self, file_path, name=None, diagram_ref=None, uid_start=1):
        super().__init__(file_path)
        self._diagram_ref = diagram_ref
        self._local_uid = uid_start
        self._name = name
        self._included_graph_ct = 0
        self._shapes = dict()
        self._connectors = dict()

        if not self.get_diagram_ref():
            self.set_diagram_ref(DiagramRef())

        if not self.get_name():
            self._parse_file_name()

        self._set_diagram_headers()

    def _set_diagram_headers(self):
        self._date_modified = f"{datetime.now(timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]}Z"
        self._diagram_id = str(uuid.uuid4()).split('-')[-1]

        self.set_grid_dx(self._diagram_ref.get_diagram_defaults('grid_dx'))
        self.set_grid_dy(self._diagram_ref.get_diagram_defaults('grid_dy'))
        self.set_grid_size(self._diagram_ref.get_diagram_defaults('grid_size'))
        self.set_page_width(self._diagram_ref.get_diagram_defaults('page_width'))
        self.set_page_height(self._diagram_ref.get_diagram_defaults('page_height'))

    def _parse_file_name(self):
        self._name = path.basename(self.get_file_path()).replace(r'\..*', '')

    def _retrieve_uid(self):
        self._local_uid += 1
        return self._local_uid - 1

    def add_shape(self, term_content, pos_x, pos_y, apply_scale=True):
        shape_uid = f"{self.get_diagram_id()}-{self._retrieve_uid()}"
        if apply_scale:
            pos_x, pos_y = self._translate_coords(pos_x, pos_y)
        draw_pos_x, draw_pos_y = pos_x, pos_y
        shape_width = self._diagram_ref.get_shape_param('rect_width')
        shape_height = self._diagram_ref.get_shape_param('rect_height')
        new_shape = Shape(shape_uid, term_content, draw_pos_x, draw_pos_y, shape_width, shape_height, self.get_diagram_ref())
        self._shapes[shape_uid] = new_shape

        return shape_uid

    def add_circle(self, term_content, pos_x, pos_y, apply_scale=True):
        shape_uid = f"{self.get_diagram_id()}-{self._retrieve_uid()}"
        if apply_scale:
            pos_x, pos_y = self._translate_coords(pos_x, pos_y)
        draw_pos_x, draw_pos_y = pos_x, pos_y
        shape_radius = self._diagram_ref.get_shape_param('circle_radius')
        new_shape = Circle(shape_uid, term_content, draw_pos_x, draw_pos_y, shape_radius, self.get_diagram_ref())
        self._shapes[shape_uid] = new_shape

        return(shape_uid)

    def add_connector(self, parent_id, child_id, rel_content, is_rank, inverted=False):
        connector_uid = f"{self.get_diagram_id()}-{self._retrieve_uid()}"
        label_connector_uid = f"{self.get_diagram_id()}-{self._retrieve_uid()}"
        new_connector = Connector(connector_uid, parent_id, child_id, label_connector_uid, rel_content, is_rank, inverted=inverted)
        self._connectors[connector_uid] = new_connector

        return connector_uid

    def add_straight_connector(self, parent_id, child_id, rel_content, start_pos_x=None, start_pos_y=None, end_pos_x=None, end_pos_y=None):
        connector_uid = f"{self.get_diagram_id()}-{self._retrieve_uid()}"
        label_connector_uid = f"{self.get_diagram_id()}-{self._retrieve_uid()}"
        new_connector = StraightConnector(connector_uid, parent_id, child_id, label_connector_uid, rel_content, start_x=start_pos_x, start_y=start_pos_y, end_x=end_pos_x, end_y=end_pos_y)
        self._connectors[connector_uid] = new_connector

        return connector_uid

    def draw(self):
        diagram_content = ""

        for connector in self._connectors.values():
            diagram_content += self._diagram_ref.get_template(connector.get_key()).substitute(connector.get_props())

        for shape in self._shapes.values():
            diagram_content += self._diagram_ref.get_template(shape.get_key()).substitute(shape.get_props())

        with open(self.get_file_path(), 'w') as write_file:
            write_content = self._diagram_ref.get_template('scaffold').substitute(self.get_props(diagram_content))
            write_file.write(write_content)

    def get_graph_count(self):
        return self._included_graph_ct

    def update_graph_count(self, add):
        self._included_graph_ct += add

    def get_diagram_ref(self):
        return self._diagram_ref

    def set_diagram_ref(self, diagram_ref):
        self._diagram_ref = diagram_ref

    def get_rank_connectors(self):
        return [connector for connector in self._connectors if connector.get_is_rank()]

    def get_props(self, diagram_content):
        return {
            'modify_date': self.get_modify_date(),
            'diagram_name': self.get_name(),
            'diagram_id': self.get_diagram_id(),
            'grid_dx': self.get_grid_dx(),
            'grid_dy': self.get_grid_dy(),
            'grid_size': self.get_grid_size(),
            'page_width': self.get_page_width(),
            'page_height': self.get_page_height(),
            'diagram_content': diagram_content
        }

    def _translate_coords(self, x_pos, y_pos, origin_x=0, origin_y=0):
        rect_width = self._diagram_ref.get_shape_param('rect_width')
        rect_height = self._diagram_ref.get_shape_param('rect_height')
        x_padding = self._diagram_ref.get_shape_param('x_padding')
        y_padding = self._diagram_ref.get_shape_param('y_padding')

        grid_x =  rect_width * 2 + x_padding
        grid_y = rect_height * 2 + y_padding

        return ((x_pos + origin_x) * grid_x, (y_pos + origin_y) * grid_y)


import argparse
from os import path

from cemento.draw_io.read_diagram import ReadDiagram
from cemento.draw_io.write_error_diagram import WriteErrorDiagram
from cemento.rdf.write_array import WriteArray

SEP_WIDTH = 75

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert draw.io diagrams into ontologies."
    )
    parser.add_argument(
        "-p",
        "--path",
        dest="file_path",
        required=True,
        help="absolute file path to diagram input. Must be a draw.io file.",
    )
    parser.add_argument(
        "-w",
        "--write_error_diagram",
        dest="should_write_error_diagram",
        default=False,
        action="store_true",
        help="whether or not to draw a separate diagram highlighting user errors.",
    )
    parser.add_argument(
        "-e",
        "--check_errors",
        default=True,
        dest="should_check_errors",
        action="store_true",
        help="whether or not to check for errors and print. True when write_error_diagram is true.",
    )
    parser.add_argument(
        "-d",
        "--do_parse",
        default=False,
        dest="do_parse",
        action="store_true",
        help="whether or not to parse the diagram as if it contained explicit names, definitions, etc.",
    )
    parser.add_argument(
        "-t",
        "--write_triples",
        default=False,
        dest="write_triples",
        action="store_true",
        help="whether or not to write an excel sheet of the triples",
    )
    parser.add_argument(
        "-f",
        "--finalize",
        dest="finalize",
        action="store_true",
        help="whether or not to finalize the diagram and generate a RDF ontology from the file. Only use this when the diagram is ready to be exported.",
    )
    args = parser.parse_args()

    should_check_errors = args.should_check_errors or args.should_write_error_diagram
    read_diagram = ReadDiagram(
        args.file_path, do_check_errors=should_check_errors, parse_terms=args.do_parse
    )

    if args.should_write_error_diagram:
        write_error_diagram = WriteErrorDiagram(read_diagram)
        write_error_diagram.add_error_highlighting()

    if should_check_errors:
        read_diagram.print_errors(sep_width=SEP_WIDTH)

    if args.do_parse:
        print(SEP_WIDTH * "=")
        print("extracted term information:", end=2 * "\n")
        term_info = read_diagram.get_terms_info()
        if term_info.empty:
            print("no term info extracted.")
        else:
            print(term_info)

    if args.write_triples:
        rels_df = read_diagram.get_relationships()
        folder_path, file_name = path.split(read_diagram.get_file_path())
        file_name, _ = file_name.split(".")

        print(SEP_WIDTH * "=")
        print("dataframe preview:", end=2 * "\n")
        if rels_df.empty:
            print("no triples generated from file.")
        else:
            print(rels_df)
            new_file_path = path.join(folder_path, f"{file_name}-triples.csv")
            rels_df.to_csv(new_file_path)
            print(f"saved to -> {new_file_path}")

        if args.do_parse and not term_info.empty:
            term_info_file_path = path.join(folder_path, f"{file_name}-term_info.csv")
            term_info.to_csv(term_info_file_path)

    if args.finalize:
        write_array = WriteArray(read_diagram)
        print(write_array.get_var_array())
        print(write_array.get_rel_array())


from os import path

import pandas as pd
from bs4 import BeautifulSoup
from defusedxml import ElementTree as ET

from cemento.draw_io.diagram import Diagram
from cemento.draw_io.diagram_errors import (
    BlankEdgeLabelError,
    BlankTermLabelError,
    CircularEdgeError,
    DisconnectedTermError,
    FloatingEdgeError,
    MissingChildEdgeError,
    MissingParentEdgeError,
    WrongFileFormatError,
)
from cemento.draw_io.diagram_ref import DiagramRef


class ReadDiagram(Diagram):

    def __init__(
        self,
        file_path,
        do_check_errors=True,
        diagram_ref=None,
        inverted_rank_arrows=False,
        parse_terms=False,
        edges=None
    ):

        if not file_path.endswith(".drawio"):
            raise WrongFileFormatError()

        super().__init__(file_path)
        self._diagram_ref = diagram_ref
        self._do_check_errors = do_check_errors
        self._inverted_rank_arrows = inverted_rank_arrows
        self._error_list = None
        self._parse_terms = parse_terms
        self._terms_info = None

        self._relationships = None

        if not self.get_diagram_ref():
            self.set_diagram_ref(DiagramRef())

        self._retrieve_diagram_headers()

        if edges is None:
            edges = self._read_edges()

        self._set_relationships(edges)

        if self._get_parse_terms():
            self._parse_entity_info()

        # retrieve file headers

    def _retrieve_diagram_headers(self):
        tree = ET.parse(self.get_file_path())
        root = tree.getroot()
        self._modify_date = root.attrib["modified"]

        diagram_tag = next(root.iter("diagram"))
        self._name = diagram_tag.attrib["name"]
        self._diagram_id = diagram_tag.attrib["id"]

        graph_model_tag = next(root.iter("mxGraphModel"))
        self.set_grid_dx(graph_model_tag.attrib["dx"])
        self.set_grid_dy(graph_model_tag.attrib["dy"])
        self.set_grid_size(int(graph_model_tag.attrib["gridSize"]))
        self.set_page_width(int(graph_model_tag.attrib["pageWidth"]))
        self.set_page_height(int(graph_model_tag.attrib["pageHeight"]))

    # convert tree based, attribute based data structure into python compatible json-like dictionary
    def _parse_elements(self):
        # parse the xml tree and set the root tag as the root
        tree = ET.parse(self.get_file_path())
        root = tree.getroot()
        root = next(tree.iter("root"))

        element = dict()
        all_cells = [
            child for child in root.findall("mxCell")
        ]  # only add base level cells

        for cell in all_cells:
            cell_attrs = dict()
            cell_attrs.update(cell.attrib)

            # add nested position attributes if any
            nested_attrs = dict()
            for subcell in cell.findall('mxGeometry'):
                nested_attrs.update(subcell.attrib)
            cell_attrs.update(nested_attrs)

            # add style attributes as data attributes
            if "style" in cell.attrib:
                style_terms = dict()
                style_tags = []
                for style_attrib_string in cell.attrib["style"].split(";"):
                    style_term = style_attrib_string.split("=")
                    if len(style_term) > 1:  # for styles with values
                        key, value = style_term
                        style_terms[key] = value
                    elif style_term[0]:  # for style tags or names
                        style_tags.append(style_term[0])
                cell_attrs.update(style_terms)
                cell_attrs["tags"] = style_tags
                del cell_attrs["style"]  # remove style to prevent redundancy

            cell_id = cell.attrib["id"]
            del cell_attrs["id"]  # remove id since it is now used as a key
            element[cell_id] = (
                cell_attrs  # set dictionary of cell key and attribute dictionary
            )

        return element

    # convert draw_io diagram into table of edges
    def _read_edges(self, save_path=None, elements=None, return_entity_ids=False):
        # classify elements in graph as terms or connectors
        if elements is None:
            elements = self._parse_elements()

        terms, rels = set(), set()
        for element, data in elements.items():
            # if the vertex attribute is 1 and edgeLabel is not in tags, it is a term (shape)
            if (
                "vertex" in data
                and data["vertex"] == "1"
                and (
                    "tags" not in data
                    or ("tags" in data and "edgeLabel" not in data["tags"])
                )
            ):
                terms.add(element)
            # if an element has an edgeLabel tag, it is a relationship (connection)
            elif "tags" in data and "edgeLabel" in data["tags"]:
                rel_val = data["value"]
                rel_id = data["parent"]
                elements[rel_id]["value"] = rel_val
                rels.add(rel_id)

        extra_rels = set()
        if self._get_check_errors():
            for element, data in elements.items():
                if "edge" in data and data["edge"] == "1":
                    extra_rels.add(element)

        black_list = set()
        if self._get_check_errors():
            self._check_errors(elements, rels, terms, extra_rels=extra_rels)
            black_list = self._error_list.keys()

        # for identified connectors, extract relationship information
        entries = []
        for rel in rels:

            # if the current relationship id has errors, skip
            if rel in black_list:
                continue

            parent_id = elements[rel]["source"]
            child_id = elements[rel]["target"]
            rel_val = ReadDiagram.clean_term(elements[rel]["value"])

            # if any of the connected terms have errors, skip
            if parent_id in black_list or child_id in black_list:
                continue

            is_rank = self._diagram_ref.is_rank(rel_val)

            # arrow conventions are inverted for rank relationships, flip assignments to conform
            if self.get_inverted_rank_arrows() and is_rank:
                temp = parent_id
                parent_id = child_id
                child_id = temp

            entries.append(
                {
                    "parent_id": parent_id,
                    "child_id": child_id,
                    "rel_id": rel,
                    "parent": ReadDiagram.clean_term(elements[parent_id]["value"]),
                    "child": ReadDiagram.clean_term(elements[child_id]["value"]),
                    "rel": rel_val,
                    "is_rank": is_rank,
                }
            )

        df = pd.DataFrame(entries)

        if save_path:
            df.to_csv(save_path)

        if return_entity_ids:
            return df, terms, rels
        return df

    def _parse_info(self, term, term_parser_titles=None):

        if not term_parser_titles:
            term_parser_titles = self.get_diagram_ref().get_term_parser_titles()

        sep = self.get_diagram_ref().get_term_parser_sep()
        # remove all the html tags in the term
        soup = BeautifulSoup(term, "html.parser")
        term_text = soup.get_text(separator=" ", strip=True)

        # if the name is defined explicitly, parse the term
        # separate string based on keyword presence
        entries = dict()
        term_breaks = [
            term_text.find(f"{key_word}{sep}")
            for key_word in term_parser_titles.values()
        ]  # first find keyword start index
        term_breaks = sorted(
            [term for term in term_breaks if term >= 0]
        )  # sort and remove non-matches
        # use break positions to split the text into a dictionary
        term_break_size = len(term_breaks)
        for i in range(term_break_size):
            start = term_breaks[i]
            end = len(term_text) if i == term_break_size - 1 else term_breaks[i + 1]
            substring = term_text[start:end]
            key, value = substring.split(
                sep, 1
            )  # only use the first occurence from left to right
            entries[key.strip()] = value.strip()

        return term_text, entries

    def _parse_entity_info(self):
        rels = self.get_relationships()
        term_parser_titles = self.get_diagram_ref().get_term_parser_titles()
        term_sep = self.get_diagram_ref().get_shape_param("term_separator")

        term_entries = []
        for row_idx, row in rels.iterrows():
            cols = ["parent", "child", "rel"]
            terms = [row[col] for col in cols]
            parent_term, child_term = None, None

            for col_idx, term in enumerate(terms):

                term_text, entries = self._parse_info(term, term_parser_titles)
                term_name = term_text  # set default name from content

                if f"{term_parser_titles['name_title']}:" in term:
                    # rename the row with the cleaned name, set term name to cleaned name
                    term_name = entries[
                        self.get_diagram_ref().get_term_parser_title("name_title")
                    ]
                    rels.at[row_idx, cols[col_idx]] = term_name

                # extract ontology from prefix
                split_term = term_name.split(term_sep, 1)
                if len(split_term) > 1:
                    entries["prefix"], entries["variable"] = split_term
                else:
                    entries["prefix"], entries["variable"] = "ex", split_term[-1]

                # assign term type if term
                entries["is_term"] = (
                    cols[col_idx] == "parent" or cols[col_idx] == "child"
                )

                # save parent term if parent
                if cols[col_idx] == "parent":
                    parent_term = entries["variable"]

                # assign parent if the term is a child
                if cols[col_idx] == "child":
                    entries["parent"] = parent_term
                    child_term = entries["variable"]

                # assign rank if the term is a relationshp
                if cols[col_idx] == "rel":
                    entries["is_data_prop"] = row["is_rank"]
                    entries["subject"] = parent_term
                    entries["object"] = child_term

                # add term as part of the dictionary
                term_entries.append(entries)

        # save all valid terms as dataframe and return
        terms_info_df = pd.DataFrame(term_entries)
        terms_info_df.dropna(how="all", inplace=True)  # only save terms with info

        self._terms_info = terms_info_df

    def _check_errors(self, elements, rels, terms, extra_rels=None):
        # check for errors if present in the diagram file. Assumes draw.io initial fidelity checks pass (file is not edited outside app)
        error_list = dict()
        all_connected_terms = set()

        if extra_rels:
            rels.update(extra_rels)

        for rel_id in rels:
            rel_errors = []
            rel_data = elements[rel_id]
            rel_content = rel_data["value"] if "value" in rel_data else ""
            connected_terms = set()

            if (
                "source" in rel_data
                and "target" in rel_data
                and rel_data["source"] == rel_data["target"]
            ):
                rel_errors.append(CircularEdgeError(rel_id, rel_content))
            elif "source" not in rel_data and "target" not in rel_data:
                rel_errors.append(FloatingEdgeError(rel_id, rel_content))
            elif "source" not in rel_data:
                rel_errors.append(MissingParentEdgeError(rel_id, rel_content))
                connected_terms.add(rel_data["target"])
            elif "target" not in rel_data:
                rel_errors.append(MissingChildEdgeError(rel_id, rel_content))
                connected_terms.add(rel_data["source"])
            else:
                connected_terms.add(rel_data["source"])
                connected_terms.add(rel_data["target"])

            all_connected_terms.update(connected_terms)

            if "value" not in rel_data or not rel_data["value"].strip():
                connected_terms = [
                    (
                        elements[term_id]["value"]
                        if "value" in elements[term_id]
                        and elements[term_id]["value"].strip()
                        else "(blank)"
                    )
                    for term_id in connected_terms
                ]
                rel_errors.append(BlankEdgeLabelError(rel_id, connected_terms))

            edge_error = {"type": "rel", "value": rel_content, "errors": rel_errors}

            if rel_errors:
                error_list[rel_id] = edge_error

        non_connected_terms = terms - all_connected_terms

        for term_id in terms:
            term_data = elements[term_id]
            term_errors = []
            term_content = term_data["value"] if "value" in term_data else ""

            if "value" not in term_data or not term_data["value"].strip():
                term_errors.append(BlankTermLabelError(rel_id))

            if term_id in non_connected_terms:
                term_errors.append(DisconnectedTermError(rel_id, term_content))

            term_error = {"type": "term", "value": term_content, "errors": term_errors}

            if term_errors:
                error_list[term_id] = term_error

        self._error_list = error_list

    def get_relationships(self):
        return self._relationships

    def _set_relationships(self, relationships):
        self._relationships = relationships

    def get_diagram_ref(self):
        return self._diagram_ref

    def set_diagram_ref(self, diagram_ref):
        self._diagram_ref = diagram_ref

    def _get_check_errors(self):
        return self._do_check_errors

    def _set_check_errors(self, do_check_errors):
        self._do_check_errors = do_check_errors

    def get_inverted_rank_arrows(self):
        return self._inverted_rank_arrows

    def _set_inverted_rank_arrows(self, inverted_rank_arrows):
        self._inverted_rank_arrows = inverted_rank_arrows

    def _get_parse_terms(self):
        return self._parse_terms

    def get_terms_info(self):
        return self._terms_info

    def get_errors(self):
        if not self._get_check_errors():
            raise ValueError(
                "Cannot retrieve error list since errors were not checked. Please make sure the argument do_check_errors is set."
            )
        return self._error_list

    def print_errors(self, sep_width=25, print_id_width=25):
        errors = self.get_errors()
        file_path = self.get_file_path()
        file_name, file_ext = path.basename(file_path).split(".")
        # print the list of errors in the console for reference
        print(sep_width * "=")
        print(
            f"Errors for {file_name}.{file_ext}\nLocated in -> {file_path}",
            end=2 * "\n",
        )
        if len(errors) > 0:
            print(f"{'Element ID'.rjust(print_id_width)}\t\tError")
            for elem_id, error_data in errors.items():
                for error in error_data["errors"]:
                    print(f"{elem_id.rjust(print_id_width)}\t->\t{error}")
        else:
            print("No errors to display.")

    @staticmethod
    def clean_term(term):
        soup = BeautifulSoup(term, "html.parser")
        term_text = soup.get_text(separator="", strip=True)
        return term_text
